library nui_web;
import 'dart:convert';
import 'dart:developer';
import 'dart:io';
import 'dart:typed_data';
import 'package:flutter/cupertino.dart';
import 'package:http/http.dart' as http;
import 'package:nui_core/busevent/nui_busevent.dart';
import 'package:nui_core/code/nui_code_util.dart';
import 'package:nui_core/data/nui_data_util.dart';
import 'package:nui_core/data/nui_ent.dart';
import 'package:nui_core/nui_core.dart';
import 'package:nui_error_handler/nui_error_handler.dart';
import 'package:nui_web/nui_connection.dart';
import 'nui_web_models.dart';

export 'package:nui_web/nui_web.dart';
export 'package:nui_web/nui_web_models.dart';
export 'package:nui_web/nui_connection.dart';

typedef HttpHeaderSetter = Map<String, String> Function();
typedef HttpExceptionWrapper = Future<NUIHttpResponse> Function();
typedef HttpAdditionalHandler = void Function(NUIHttpResponse);

abstract class NUIWebInterface{
  Future<NUIHttpResponse<T>> doPOST<T extends NUIEnt>({@required String url, Map<String, String> headers, NUIEnt bodyEnt, Map<String, String> params, Map<String, dynamic> body, Encoding encoding, String pathVariable, NUIEntMapper<T> entity, bool storeTemp});
  Future<NUIHttpResponse<T>> doPUT<T extends NUIEnt>({@required String url, Map<String, String> headers, NUIEnt bodyEnt, Map<String, String> params, Map<String, dynamic> body, Encoding encoding, String pathVariable, NUIEntMapper<T> entity, bool storeTemp});
  Future<NUIHttpResponse<T>> doPATCH<T extends NUIEnt>({@required String url, Map<String, String> headers, NUIEnt bodyEnt, Map<String, String> params, Map<String, dynamic> body, Encoding encoding, String pathVariable, NUIEntMapper<T> entity, bool storeTemp});
  Future<NUIHttpResponse<T>> doDELETE<T extends NUIEnt>({@required String url, Map<String, String> headers, Encoding encoding, Map<String, String> params, String pathVariable, NUIEntMapper<T> entity, bool storeTemp});
  Future<NUIHttpResponse<T>> doGET<T extends NUIEnt>({@required String url, Map<String, String> headers, String pathVariable, Map<String, String> params, NUIEntMapper<T> entity, bool storeTemp});
}

class NUIWeb implements NUIWebInterface{
  static const _NUIWEB_MODULE = "NUIWeb";

  static const int ERROR_INTERCEPTOR = 1090;

  NUIWeb._builder(NUIWebBuilder builder) : _baseUrl = builder._baseUrl, _preHandler = builder._preHandler, _postHandlerDelay = builder._postHandlerDelay,  _postHandler = builder._postHandler, _headers = builder._headers, _statusMessages = builder._statusMessages ?? Map(), _interceptors = builder._interceptors, _errorHandler = builder._errorHandler, _connectivityListener = builder._connectivityListener, _module = isNullOrEmpty(builder._module) ? _defaultModule : builder._module{
    _webMap[_module] = this;
    NUISharedPrefBuilder().build();
    NUIBusEventBuilder().module(_getBusEventModule()).build();
    NUIConnectionUtil.subscribeConnectivityChange((result, hasInternet) {
      _getBusEvent().send<bool>(NUIBusEventData<bool>(data: hasInternet, sender: _getBusEventModule(), code: NUIConnectionUtil.CONNECTIVITY_CHANNEL), channel: NUIConnectionUtil.CONNECTIVITY_CHANNEL);
      if(_connectivityListener != null) {
        _connectivityListener(result, hasInternet);
      }
    });
  }

  final Map<int, String> _statusMessages;
  final String _module;
  final String _baseUrl;
  final HttpAdditionalHandler _preHandler;
  final HttpAdditionalHandler _postHandler;
  final Duration _postHandlerDelay;
  final ConnectivityChangeListener _connectivityListener;
  final HttpHeaderSetter _headers;
  final NUIErrorHandler _errorHandler;
  final List<NUIWebInterceptor> _interceptors;
  static Map<String, NUIWeb> _webMap = Map();
  static const String _defaultModule = "NUIWebDefaultModule";

  static Future<Uint8List> downloadFromURL(String url) async{
    try {
      http.Client client = new http.Client();
      var req = await client.get(Uri.parse(url));
      var bytes = req.bodyBytes;
      return bytes;
    }catch(e){
      return null;
    }
  }

  static NUIWeb get({String module}){
    NUIWeb instance = _webMap[isNullOrEmpty(module) ? _defaultModule : module];
    if(instance == null) throw Exception("NUIWeb -> NUIWeb instance has not been initialized with builder");
    return instance;
  }

  NUIBusEvent _getBusEvent(){
    return NUIBusEvent.get(module: _getBusEventModule());
  }

  void sendBusEvent<T>(NUIBusEventData<T> data, {String channel: NUIConnectionUtil.CONNECTIVITY_CHANNEL}){
    _getBusEvent().send<T>(data, channel: channel);
  }

  void subscribeBusEvent<T>(NUIBusEventListener listener, {String channel: NUIConnectionUtil.CONNECTIVITY_CHANNEL}){
    _getBusEvent().listen(listener, channel: channel);
  }

  void disposeBusEvent(NUIBusEventListener listener, {String channel: NUIConnectionUtil.CONNECTIVITY_CHANNEL}){
    _getBusEvent().dispose(listener, channel: channel);
  }

  String _getBusEventModule(){
    return "NUIWebBusEvent_" + (_module ?? _defaultModule);
  }

  Future<NUIStandardHttpResponse> _processInterceptors(String url, NUIWebMethod method, Map<String, String> headers, Map<String, dynamic> body) async{
    //No interceptors, so return true
    if(isNullOrEmpty(_interceptors)) return null;

    for(NUIWebInterceptor interceptor in _interceptors){
      final result = await interceptor.intercept(url, method, dynamic, headers, body);
      if(result.interceptResponse == true) return result;
    }
    return null;
  }

  String _getProcessedUrl(String url, String pathVariable, Map<String, String> params){
    String webUrl = isNullOrEmpty(_baseUrl) ? url : (_baseUrl[_baseUrl.length - 1] != '/' ? (_baseUrl + "/" + url) : (_baseUrl + url));
    if(!isNullOrEmpty(pathVariable)){
      if(webUrl[webUrl.length - 1] != '/') {
        webUrl = webUrl + "/" + pathVariable;
      }
      else{
        webUrl = webUrl + pathVariable;
      }
    }
    webUrl =  webUrl + _getPathParams(params);
    logNUI(NUIWeb._NUIWEB_MODULE, "URL : $webUrl");
    return webUrl;
  }

  dynamic _getContentBody(NUIEnt bodyEnt, Map<String, dynamic> body){
    final bodyContent = bodyEnt != null ? jsonEncode(bodyEnt) : jsonEncode(body);

    if(bodyContent != null) {
      final prettyString = JsonEncoder.withIndent("   ").convert(bodyContent);
      logNUI(NUIWeb._NUIWEB_MODULE, "Request body :\n$prettyString");
    }

    return bodyContent;
  }

  void _logWeb(String message){
    logNUI(NUIWeb._NUIWEB_MODULE, "$message");
  }

  void _preHandle<T>(NUIHttpResponse<T> response){
    if(_preHandler != null) _preHandler(response);
  }

  void _postHandle<T>(NUIHttpResponse<T> response){
    if(_postHandler != null){
      NUIAsync.delay(_postHandlerDelay, () {
        _postHandler(response);
      });
    }
  }

  String _getPathParams(Map<String, String> params){
    if(params == null || params.isEmpty){
      return "";
    }
    String pathParam = "";
    for(String key in params.keys){
      pathParam = pathParam + (pathParam.isEmpty ? "?$key=${params[key]}" : "&$key=${params[key]}");
    }
    return pathParam;
  }

  Map<String, String> _getHeaders(Map<String, String> additionalHeaders){
    if(_headers == null) {
      _logWeb("Headers : " + additionalHeaders.toString());
      return additionalHeaders;
    }
    if(additionalHeaders == null) {
      _logWeb("Headers : " + _headers().toString());
      return _headers();
    }

    Map<String, String> headers = _headers();
    for(String key in additionalHeaders.keys){
      headers[key] = additionalHeaders[key];
    }
    if(!isNullOrEmpty(headers)){
      _logWeb("Headers : " + headers.toString());
    }
    return headers;
  }

  Future<NUIHttpResponse<T>> wrapException<T extends NUIEnt>(HttpExceptionWrapper task) async{
    try {
      return task();
    }catch(e){
      _logWeb("Encountered an exception : ${e.toString()}");
      return NUIHttpResponse(success: false, code: HttpStatus.internalServerError, message: e.toString());
    }
  }

  Future<bool> _checkInternetConnection() async{
    return NUIConnectionUtil.hasConnection();
  }

  void _storeCacheResponse({String url, String pathVariable, http.Response response}){
    if(response.statusCode != HttpStatus.ok) return;
    final fullUrl = _getProcessedUrl(url, pathVariable, null);
    final cacheResponse = NUIHttpCacheResponse(response: response.body, headers: response.headers);
    NUISharedPref.get().save(fullUrl, cacheResponse);
  }

  NUIHttpCacheResponse getCacheResponse({String url, String pathVariable}){
    final fullUrl = _getProcessedUrl(url, pathVariable, null);
    return NUISharedPref.get().quickGetObject(fullUrl, NUIHttpCacheResponseEntity());
  }

  NUIHttpResponse<D> _processResponse<D>(http.Response response, NUIEntMapper entity){
    _logWeb("Response body : " + response?.body);
    if(response.statusCode == HttpStatus.ok){
      if(entity != null){
        final resp = NUIHttpResponse<D>(
            success: true,
            code: response.statusCode,
            message: _statusMessages[response.statusCode] ?? "Http status OK",
            responseBody: jsonDecode(response.body),
            response: response.body,
            headers: response.headers
        );

        final bodyMap = jsonDecode(response.body);
        if(bodyMap is List){
          resp.listBody = asListOfObjects(entity, bodyMap);
        }
        else{
          resp.body = entity.toEntity(bodyMap);
        }
        return resp;
      }
      return NUIHttpResponse<D>(
          success: true,
          code: response.statusCode,
          message: _statusMessages[response.statusCode] ?? "Http status OK",
          responseBody: jsonDecode(response.body),
          response: response.body,
          headers: response.headers
      );
    }
    else {
      _processError(response.statusCode?.toString());
      Map<String, dynamic> decodedBody;
      String errorMessage;
      try{
        decodedBody = jsonDecode(response.body);
      }catch(e){
        logNUI("NUIWeb", "Failed to decode failure response with error : $e");
        //Error message might be the response body as string itself
        errorMessage = response.body;
      }
      return NUIHttpResponse<D>(
          success: false,
          code: response.statusCode,
          message: _statusMessages[response.statusCode] ?? errorMessage ?? "No message specified",
          response: response.body,
          responseBody: decodedBody,
          headers: response.headers
      );
    }
  }

  void _processError(String code){
    if(_errorHandler != null){
      final context = NUINavigator.getLastContext();
      if(context != null) {
        _errorHandler.processError(context, code, errorType: NUIErrorType.HTTP);
      }
    }
  }

  Future<NUIHttpResponse<T>> _preRequestChecking<T extends NUIEnt>(String url, NUIWebMethod method, Map<String, String> headers, Map<String, dynamic> body) async{
    final hasConnection = await _checkInternetConnection();
    if(hasConnection) {
      //Has connection, now run interceptors
      final interceptResult = await _processInterceptors(url, method, headers, body);
      if(interceptResult != null){
        return NUIHttpResponse<T>(success: interceptResult.success, code: interceptResult.code ?? NUIWeb.ERROR_INTERCEPTOR, message: interceptResult.message, responseBody: interceptResult.data);
      }
      return null;
    }
    else{
      _processError(NUIErrorCode.NO_INTERNET_CONNECTION.value().toString());
      return NUIHttpResponse<T>(success: false, code: NUIErrorCode.NO_INTERNET_CONNECTION.value(), message: "No internet connection found");
    }
  }

  @override
  Future<NUIHttpResponse<T>> doPOST<T extends NUIEnt>({String url, Map<String, String> headers, NUIEnt bodyEnt, Map<String, dynamic> body, Map<String, String> params, Encoding encoding, String pathVariable, NUIEntMapper<T> entity, bool storeTemp: false}) async{
    final preprocess = await _preRequestChecking<T>(url, NUIWebMethod.POST, headers, body);
    if(preprocess != null) return preprocess;
    final targetUrl = _getProcessedUrl(url, pathVariable, params);
    final contentBody = _getContentBody(bodyEnt, body);
    NUIHttpResponse<T> response = await wrapException<T>(() async{
      http.Response response = await http.post(
          Uri.parse(targetUrl),
          headers: _getHeaders(headers),
          body: contentBody,
          encoding: encoding
      );
      if(storeTemp == true) _storeCacheResponse(url: targetUrl, pathVariable: pathVariable, response: response);
      return _processResponse<T>(response, entity);
    });
    _preHandle<T>(response);
    _postHandle(response);
    return response;
  }

  @override
  Future<NUIHttpResponse<T>> doPUT<T extends NUIEnt>({String url, Map<String, String> headers, NUIEnt bodyEnt, Map<String, String> params, Map<String, dynamic> body, Encoding encoding, String pathVariable, NUIEntMapper<T> entity, bool storeTemp: false}) async{
    final preprocess = await _preRequestChecking<T>(url, NUIWebMethod.PUT, headers, body);
    if(preprocess != null) return preprocess;
    final targetUrl = _getProcessedUrl(url, pathVariable, params);
    final contentBody = _getContentBody(bodyEnt, body);
    NUIHttpResponse<T> response = await wrapException<T>(() async{
      http.Response response = await http.put(
          Uri.parse(targetUrl),
          headers: _getHeaders(headers),
          body: contentBody,
          encoding: encoding
      );
      if(storeTemp == true) _storeCacheResponse(url: targetUrl, pathVariable: pathVariable, response: response);
      return _processResponse<T>(response, entity);
    });
    _preHandle<T>(response);
    _postHandle(response);
    return response;
  }

  @override
  Future<NUIHttpResponse<T>> doPATCH<T extends NUIEnt>({String url, Map<String, String> headers, NUIEnt bodyEnt, Map<String, String> params, Map<String, dynamic> body, Encoding encoding, String pathVariable, NUIEntMapper<T> entity, bool storeTemp: false}) async{
    final preprocess = await _preRequestChecking<T>(url, NUIWebMethod.PATCH, headers, body);
    if(preprocess != null) return preprocess;
    final targetUrl = _getProcessedUrl(url, pathVariable, params);
    final contentBody = _getContentBody(bodyEnt, body);
    NUIHttpResponse<T> response = await wrapException<T>(() async{
      http.Response response = await http.patch(
          Uri.parse(targetUrl),
          headers: _getHeaders(headers),
          body: contentBody,
          encoding: encoding
      );
      if(storeTemp == true) _storeCacheResponse(url: targetUrl, pathVariable: pathVariable, response: response);
      return _processResponse<T>(response, entity);
    });
    _preHandle<T>(response);
    _postHandle(response);
    return response;
  }

  @override
  Future<NUIHttpResponse<T>> doDELETE<T extends NUIEnt>({String url, Map<String, String> headers, Map<String, String> params, Encoding encoding, String pathVariable, NUIEntMapper<T> entity, bool storeTemp: false}) async{
    final preprocess = await _preRequestChecking<T>(url, NUIWebMethod.DELETE, headers, null);
    if(preprocess != null) return preprocess;
    final targetUrl = _getProcessedUrl(url, pathVariable, params);
    NUIHttpResponse<T> response = await wrapException<T>(() async{
      http.Response response = await http.delete(
          Uri.parse(targetUrl),
          headers: _getHeaders(headers)
      );
      if(storeTemp == true) _storeCacheResponse(url: targetUrl, pathVariable: pathVariable, response: response);
      return _processResponse<T>(response, entity);
    });
    _preHandle<T>(response);
    _postHandle(response);
    return response;
  }

  @override
  Future<NUIHttpResponse<T>> doGET<T extends NUIEnt>({String url, Map<String, String> headers, String pathVariable, Map<String, String> params, NUIEntMapper<T> entity, bool storeTemp}) async{
    final preprocess = await _preRequestChecking<T>(url, NUIWebMethod.GET, headers, null);
    if(preprocess != null) return preprocess;
    final targetUrl = _getProcessedUrl(url, pathVariable, params);
    NUIHttpResponse<T> response = await wrapException<T>(() async{
      http.Response response = await http.get(
        Uri.parse(targetUrl),
        headers: _getHeaders(headers),
      );
      if(storeTemp == true) _storeCacheResponse(url: targetUrl, pathVariable: pathVariable, response: response);
      return _processResponse<T>(response, entity);
    });
    _preHandle<T>(response);
    _postHandle(response);
    return response;
  }
}

class NUIWebBuilder{
  Map<int, String> _statusMessages;
  String _baseUrl;
  HttpHeaderSetter _headers;
  String _module;
  HttpAdditionalHandler _preHandler;
  HttpAdditionalHandler _postHandler;
  Duration _postHandlerDelay;
  ConnectivityChangeListener _connectivityListener;
  ErrorHandlerGetter _errorHandlerBuilder;
  NUIErrorHandler _errorHandler;
  List<NUIWebInterceptor> _interceptors = List();

  NUIWebBuilder module(String module){
    _module = module;
    return this;
  }
  NUIWebBuilder errorHandler(ErrorHandlerGetter errorHandler){
    _errorHandlerBuilder = errorHandler;
    return this;
  }
  NUIWebBuilder preHandler(HttpAdditionalHandler handler){
    _preHandler = handler;
    return this;
  }
  NUIWebBuilder addInterceptor(NUIWebInterceptor interceptor){
    _interceptors.add(interceptor);
    return this;
  }
  NUIWebBuilder postHandler(HttpAdditionalHandler handler, {Duration delay}){
    _postHandler = handler;
    _postHandlerDelay = delay ?? Duration(milliseconds: 500);
    return this;
  }
  NUIWebBuilder connectivityListener(ConnectivityChangeListener listener){
    this._connectivityListener = listener;
    return this;
  }
  NUIWebBuilder statusMessages(Map<int, String> messages){
    _statusMessages = messages;
    return this;
  }
  NUIWebBuilder baseURL(String url){
    _baseUrl = url;
    return this;
  }
  NUIWebBuilder headers(HttpHeaderSetter headers){
    _headers = headers;
    return this;
  }
  Future<NUIWeb> build() async{
    if(_errorHandlerBuilder != null){
      _errorHandler = await _errorHandlerBuilder().build();
    }
    final web = NUIWeb._builder(this);
    logNUI(NUIWeb._NUIWEB_MODULE, "Initialization successful");
    return web;
  }
}
